#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
import base64
import hashlib
import hmac
import os
import random
import sys
from datetime import datetime
from os import urandom

# import interfaces
from pyams_security.interfaces import ISecurityManager, IUsersFolderPlugin, ILocalUser, IPrincipalInfo
from pyams_security.interfaces.notification import INotificationSettings
from pyramid_chameleon.interfaces import IChameleonTranslate
from pyramid_mailer.interfaces import IMailer
from zope.component.interfaces import ISite
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.password.interfaces import IPasswordManager
from zope.schema.interfaces import IVocabularyRegistry

# import packages
from chameleon.zpt.template import PageTemplateFile
from persistent import Persistent
from pyams_security.principal import PrincipalInfo
from pyams_utils.adapter import adapter_config
from pyams_utils.html import html_to_text
from pyams_utils.registry import query_utility, get_utility
from pyams_utils.request import check_request
from pyams_utils.traversing import get_parent
from pyramid.events import subscriber
from pyramid_mailer.message import Message, Attachment
from zope.container.contained import Contained
from zope.container.folder import Folder
from zope.interface import implementer, provider, Invalid
from zope.schema.fieldproperty import FieldProperty
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm, getVocabularyRegistry

from pyams_security import _


@implementer(ILocalUser)
class User(Persistent, Contained):
    """Local user persistent class"""

    login = FieldProperty(ILocalUser['login'])
    email = FieldProperty(ILocalUser['email'])
    firstname = FieldProperty(ILocalUser['firstname'])
    lastname = FieldProperty(ILocalUser['lastname'])
    company_name = FieldProperty(ILocalUser['company_name'])
    _password = FieldProperty(ILocalUser['password'])
    _password_salt = None
    password_manager = FieldProperty(ILocalUser['password_manager'])
    wait_confirmation = FieldProperty(ILocalUser['wait_confirmation'])
    self_registered = FieldProperty(ILocalUser['self_registered'])
    activation_secret = FieldProperty(ILocalUser['activation_secret'])
    activation_hash = FieldProperty(ILocalUser['activation_hash'])
    activation_date = FieldProperty(ILocalUser['activation_date'])
    activated = FieldProperty(ILocalUser['activated'])

    @property
    def title(self):
        return '{firstname} {lastname}'.format(firstname=self.firstname,
                                               lastname=self.lastname)

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        self._password_salt = urandom(4)
        manager = get_utility(IPasswordManager, name=self.password_manager)
        if self.password_manager == 'Plain Text':
            self._password = manager.encodePassword(value)
        else:
            self._password = manager.encodePassword(value, salt=self._password_salt)

    def check_password(self, password):
        if not self.activated:
            return False
        manager = query_utility(IPasswordManager, name=self.password_manager)
        if manager is None:
            return False
        return manager.checkPassword(self.password, password)

    def generate_secret(self):
        seed = self.activation_secret = '-'.join((hex(random.randint(0, sys.maxsize))[2:] for i in range(5)))
        secret = hmac.new(self.password, self.login.encode(), digestmod=hashlib.sha256)
        secret.update(seed.encode())
        self.activation_hash = base64.b32encode(secret.digest()).decode()

    def check_activation(self, hash, login, password):
        if self.self_registered:
            # If principal was registered by it's own, we check activation hash
            # with given login and password
            manager = get_utility(IPasswordManager, name=self.password_manager)
            password = manager.encodePassword(password, salt=self._password_salt)
            secret = hmac.new(password, login.encode(), digestmod=hashlib.sha256)
            secret.update(self.activation_secret.encode())
            activation_hash = base64.b32encode(secret.digest()).decode()
            if hash != activation_hash:
                raise Invalid(_("Can't activate profile with given params!"))
        else:
            # If principal was registered by a site manager, just check that
            # hash is matching stored one and update user password...
            if hash != self.activation_hash:
                raise Invalid(_("Can't activate profile with given params!"))
            self.password = password
        self.activation_date = datetime.utcnow()
        self.activated = True


@adapter_config(context=ILocalUser, provides=IPrincipalInfo)
def UserPrincipalInfoAdapter(user):
    """User principal info adapter"""
    return PrincipalInfo(id="{prefix}:{login}".format(prefix=user.__parent__.prefix,
                                                      login=user.login),
                         title=user.title)


@implementer(IUsersFolderPlugin)
class UsersFolder(Folder):
    """Local users folder"""

    prefix = FieldProperty(IUsersFolderPlugin['prefix'])
    title = FieldProperty(IUsersFolderPlugin['title'])
    enabled = FieldProperty(IUsersFolderPlugin['enabled'])

    def authenticate(self, credentials, request):
        if not self.enabled:
            return None
        attrs = credentials.attributes
        login = attrs.get('login')
        principal = self.get(login)
        if principal is not None:
            password = attrs.get('password')
            if principal.check_password(password):
                return "{prefix}:{login}".format(prefix=self.prefix,
                                                 login=principal.login)

    def check_login(self, login):
        if not login:
            return False
        return login not in self

    def get_principal(self, principal_id, info=True):
        if not self.enabled:
            return None
        if not principal_id.startswith(self.prefix + ':'):
            return None
        prefix, login = principal_id.split(':', 1)
        user = self.get(login)
        if user is not None:
            if info:
                return PrincipalInfo(id='{prefix}:{login}'.format(prefix=self.prefix,
                                                                  login=user.login),
                                     title=user.title)
            else:
                return user

    def get_all_principals(self, principal_id):
        if not self.enabled:
            return set()
        if self.get_principal(principal_id) is not None:
            return {principal_id}
        return set()

    def find_principals(self, query):
        if not self.enabled:
            raise StopIteration
        # TODO: use inner text catalog for more efficient search?
        if not query:
            return None
        query = query.lower()
        for user in self.values():
            if (query == user.login or
                    query in user.firstname.lower() or
                    query in user.lastname.lower() or
                    query in user.email.lower()):
                yield PrincipalInfo(id='{prefix}:{login}'.format(prefix=self.prefix,
                                                                 login=user.login),
                                    title='{title} <{email}>'.format(title=user.title,
                                                                     email=user.email))

    def get_search_results(self, data):
        # TODO: use inner text catalog for more efficient search?
        query = data.get('query')
        if not query:
            return ()
        query = query.lower()
        for user in self.values():
            if (query == user.login or
                    query in user.firstname.lower() or
                    query in user.lastname.lower()):
                yield user


@provider(IVocabularyRegistry)
class UsersFolderVocabulary(SimpleVocabulary):
    """'PyAMS users folders' vocabulary"""

    def __init__(self, *args, **kwargs):
        terms = []
        manager = query_utility(ISecurityManager)
        if manager is not None:
            for name, plugin in manager.items():
                if IUsersFolderPlugin.providedBy(plugin):
                    terms.append(SimpleTerm(name, title=plugin.title))
        super(UsersFolderVocabulary, self).__init__(terms)

getVocabularyRegistry().register('PyAMS users folders', UsersFolderVocabulary)


@subscriber(IObjectAddedEvent, context_selector=ILocalUser)
def handle_new_local_user(event):
    """Send a confirmation message when a new user is recorded"""
    security = query_utility(ISecurityManager)
    settings = INotificationSettings(security)
    mailer = settings.mailer
    if mailer is None:
        return
    if isinstance(mailer, str):
        mailer = query_utility(IMailer, name=mailer)
        if mailer is None:
            return
    request = check_request()
    translate = request.localizer.translate
    user = event.object
    if user.self_registered:
        template_name = 'register-message.pt'
    elif user.wait_confirmation:
        template_name = 'register-info.pt'
    else:
        template_name = None
    if template_name is not None:
        template = PageTemplateFile(os.path.join(os.path.split(__file__)[0], 'templates', template_name),
                                    translate=query_utility(IChameleonTranslate))
        site = get_parent(request.context, ISite)
        html_body = template(request=request, context=user, site=site, settings=settings)
        message = Message(subject=translate(_("{prefix}Please confirm registration"))
                            .format(prefix="{prefix} ".format(prefix=settings.subject_prefix)
                                           if settings.subject_prefix else ''),
                          sender='{name} <{email}>'.format(name=settings.sender_name,
                                                           email=settings.sender_email),
                          recipients=("{firstname} {lastname} <{email}>".format(firstname=user.firstname,
                                                                                lastname=user.lastname,
                                                                                email=user.email), ),
                          html=Attachment(data=html_body,
                                          content_type='text/html; charset=utf-8',
                                          disposition='inline',
                                          transfer_encoding='quoted-printable'),
                          body=Attachment(data=html_to_text(html_body),
                                          content_type='text/plain; charset=utf-8',
                                          disposition='inline',
                                          transfer_encoding='quoted-printable'))
        mailer.send(message)
